/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::ExampleClient
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once

#include <experimental/memory>
#include <string>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/services/tasks/RunningTask.h>

#include <RobotAPI/libraries/armem/client/forward_declarations.h>
#include <RobotAPI/libraries/armem/client/plugins/ListeningPluginUser.h>
#include <RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>

#include "armarx/navigation/memory/client/costmap/Reader.h"
#include <armarx/navigation/client.h>

namespace armarx::navigation::components::example_client
{

    enum class Mode
    {
        Standard,
        Complex,
        PointToPoint,
        UpdateNavigator,
        WanderAround
    };

    /**
     * @defgroup Component-ExampleClient ExampleClient
     * @ingroup armarx_navigation-Components
     * A description of the component ExampleClient.
     *
     * @class ExampleClient
     * @ingroup Component-ExampleClient
     * @brief Brief description of class ExampleClient.
     *
     * Detailed description of class ExampleClient.
     */
    class Component :
        virtual public armarx::Component,
        virtual public armarx::navigation::client::ComponentPluginUser,
        virtual public armem::ListeningClientPluginUser
    {

    public:
        Component();

        ~Component() override;

        static std::string GetDefaultName();

    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;

        std::string getDefaultName() const override;

        void run();

        void exampleNavigation();
        void exampleNavigationComplex();

        /**
         * @brief demonstrates point to point movement.
         * 
         * The robot will move along a rectange in the following way:
         * 
         * (1) move to the right 
         * (2) move forward and turn 90° to the left 
         * (3) move left and turn back 90° to the right 
         * (4) backward (this is the initial pose)
         * 
         */
        void exampleNavigationPointToPoint();

        void exampleNavigationUpdateNavigator();

        void exampleNavigationWanderAround();

    private:
        void goalReached();

        struct
        {
            std::string robotName;
            float relativeMovement = 200; // [mm]

            Mode mode = Mode::Standard;
        } properties;

        armarx::RunningTask<Component>::pointer_type task;

        template <typename T>
        using rw_plugin = armem::client::plugins::ReaderWriterPlugin<T>;

        template <typename T>
        using obs_ptr = std::experimental::observer_ptr<T>;

        obs_ptr<rw_plugin<armem::robot_state::VirtualRobotReader>> virtualRobotReaderPlugin =
            nullptr;

        // for exampleNavigationWanderAround
        obs_ptr<rw_plugin<memory::client::costmap::Reader>> costmapReaderPlugin = nullptr;
    };

} // namespace armarx::navigation::components::example_client
