/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <functional>
#include <optional>
#include <string>
#include <vector>

#include <Eigen/Core>

#include <VirtualRobot/VirtualRobot.h>

#include <armarx/navigation/algorithms/Costmap.h>
#include <armarx/navigation/algorithms/Room.h>
#include <armarx/navigation/algorithms/types.h>

namespace armarx::navigation::algorithms
{

    class CostmapBuilder
    {
    public:
        struct CostmapBuilderParams
        {
            bool restrictToRooms = false;
            Eigen::Vector3f collisionModelScaleFactor = Eigen::Vector3f::Ones();
        };

        CostmapBuilder(const VirtualRobot::RobotPtr& robot,
                       const VirtualRobot::SceneObjectSetPtr& obstacles,
                       const std::vector<VirtualRobot::RobotPtr>& articulatedObjects,
                       const std::vector<Room>& rooms,
                       const Costmap::Parameters& parameters,
                       const std::string& robotCollisonModelName,
                       const CostmapBuilderParams& builderParameters);

        Costmap create(const SceneBounds& init = SceneBounds());
        Costmap extend(Costmap costmap);


        static Eigen::MatrixXf createUniformGrid(const SceneBounds& sceneBounds,
                                                 const Costmap::Parameters& parameters);

        static void initializeMask(Costmap& costmap);

    private:
        float computeCost(const Costmap::Position& position,
                          VirtualRobot::RobotPtr& collisionRobot,
                          const VirtualRobot::CollisionModelPtr& robotCollisionModel);

        void fillGridCosts(Costmap& costmap);
        void extendGridCosts(Costmap& costmap, const std::vector<Room>& rooms);

        // fn: collisionRobot, robotCollisionModel, mask, costs, index -> costs
        // expects mask calculation based on costs after this function
        void applyFnToCostmap(Costmap& costmap,
                              const std::function<float(VirtualRobot::RobotPtr&,
                                                        VirtualRobot::CollisionModelPtr&,
                                                        const std::optional<float>,
                                                        const float,
                                                        const Costmap::Index&)>& fn);

        void initializeEmptyMask(Costmap& costmap);

        const VirtualRobot::RobotPtr robot;
        const VirtualRobot::SceneObjectSetPtr obstacles;
        const std::vector<VirtualRobot::RobotPtr> articulatedObjects;
        const std::vector<Room> rooms;
        const Costmap::Parameters parameters;
        const std::string robotCollisionModelName;

        const CostmapBuilderParams builderParameters;
    };

} // namespace armarx::navigation::algorithms
