/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::ArmarXObjects::distance_to_obstacle_costmap_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once

#include <mutex>
#include <optional>
#include <string>
#include <vector>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>

#include <RobotAPI/libraries/ArmarXObjects/plugins/ObjectPoseClientPlugin.h>
#include <RobotAPI/libraries/armem/client/plugins/ListeningPluginUser.h>
#include <RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem_objects/client/articulated_object/ArticulatedObjectReader.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>
#include <RobotAPI/libraries/armem_robot_state/types.h>

#include <armarx/navigation/algorithms/Costmap.h>
#include <armarx/navigation/algorithms/CostmapBuilder.h>
#include <armarx/navigation/components/distance_to_obstacle_costmap_provider/ComponentInterface.h>
#include <armarx/navigation/memory/client/costmap/Reader.h>
#include <armarx/navigation/memory/client/costmap/Writer.h>
#include <armarx/navigation/memory/client/rooms/Reader.h>

namespace armarx::navigation::components::distance_to_obstacle_costmap_provider
{

    class Component :
        virtual public armarx::Component,
        virtual public ObjectPoseClientPluginUser,
        virtual public armarx::navigation::components::distance_to_obstacle_costmap_provider::
            ComponentInterface,
        virtual public armarx::armem::ListeningClientPluginUser
    // , virtual public armarx::DebugObserverComponentPluginUser
    // , virtual public armarx::LightweightRemoteGuiComponentPluginUser
    // , virtual public armarx::ArVizComponentPluginUser
    {
    public:
        Component();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        /// Get the component's default name.
        static std::string GetDefaultName();


    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;


        /* (Requires armarx::LightweightRemoteGuiComponentPluginUser.)
        /// This function should be called once in onConnect() or when you
        /// need to re-create the Remote GUI tab.
        void createRemoteGuiTab();

        /// After calling `RemoteGui_startRunningTask`, this function is
        /// called periodically in a separate thread. If you update variables,
        /// make sure to synchronize access to them.
        void RemoteGui_update() override;
        */

        bool createAndStoreCostmap();

        void processObjectInstance(const armem::MemoryID& id,
                                   const std::vector<armem::MemoryID>& snapshotIDs);

    private:
        // Private methods go here.

        // Forward declare `Properties` if you used it before its defined.
        // struct Properties;

        /* (Requires the armarx::ArVizComponentPluginUser.)
        /// Draw some boxes in ArViz.
        void drawBoxes(const Properties& p, viz::Client& arviz);
        */


        static const std::string defaultName;

        // Private member variables go here.


        /// Properties shown in the Scenario GUI.
        struct Properties
        {
            std::string robotName;
            std::string primitiveModelIds =
                "navigation,urdf_col"; // intentionally not using empty default model
            std::string colModel = "Platform-navigation-colmodel";

            algorithms::Costmap::Parameters costmapParams;
            algorithms::CostmapBuilder::CostmapBuilderParams costmapBuilderParams;

            std::string costmapToExtend;
            bool buildSecondCostmapWithSmallerMargins = true;
            std::string costmapSmallerMarginsToExtend;
        };

        Properties properties;
        /* Use a mutex if you access variables from different threads
         * (e.g. ice functions and RemoteGui_update()).
        std::mutex propertiesMutex;
        */


        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        /// Tab shown in the Remote GUI.
        struct RemoteGuiTab : armarx::RemoteGui::Client::Tab
        {
            armarx::RemoteGui::Client::LineEdit boxLayerName;
            armarx::RemoteGui::Client::IntSpinBox numBoxes;

            armarx::RemoteGui::Client::Button drawBoxes;
        };
        RemoteGuiTab tab;
        */


        /* (Requires the armarx::ArVizComponentPluginUser.)
         * When used from different threads, an ArViz client needs to be synchronized.
        /// Protects the arviz client inherited from the ArViz plugin.
        std::mutex arvizMutex;
        */
        std::optional<armem::robot_state::description::RobotDescription> robotDescription;
        armem::client::plugins::ReaderWriterPlugin<armem::robot_state::VirtualRobotReader>*
            virtualRobotReaderPlugin = nullptr;


        armem::client::plugins::ReaderWriterPlugin<
            armarx::armem::articulated_object::ArticulatedObjectReader>*
            articulatedObjectReaderPlugin = nullptr;


        armem::client::plugins::ReaderWriterPlugin<memory::client::costmap::Reader>*
            costmapReaderPlugin = nullptr;

        armem::client::plugins::ReaderWriterPlugin<memory::client::costmap::Writer>*
            costmapWriterPlugin = nullptr;


        armem::client::plugins::ReaderWriterPlugin<memory::client::rooms::Reader>*
            roomReaderPlugin = nullptr;


        std::mutex createCostmapMtx;
    };

} // namespace armarx::navigation::components::distance_to_obstacle_costmap_provider
