/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <optional>
#include <vector>

#include <VirtualRobot/MathTools.h>

#include <RobotAPI/interface/units/LaserScannerUnit.h>
#include <RobotAPI/libraries/armem_laser_scans/types.h>

#include <armarx/navigation/components/laser_scanner_feature_extraction/ChainApproximation.h>
#include <armarx/navigation/memory/types.h>

#include "EnclosingEllipsoid.h"
#include "ScanClustering.h"

namespace armarx::navigation::components::laser_scanner_feature_extraction
{
    using Circle = memory::Circle;

    struct Features
    {
        using Points = std::vector<Eigen::Vector2f>;
        using Chain = Points;

        std::optional<VirtualRobot::MathTools::ConvexHull2D> convexHull;

        std::optional<Circle> circle;
        std::optional<Ellipsoid> ellipsoid;

        std::optional<Chain> chain;

        Points points;
        std::vector<Ellipsoid> linesAsEllipsoids(float axisLength) const;
    };

    class FeatureExtractor
    {
    public:
        using Points = std::vector<Eigen::Vector2f>;

        FeatureExtractor(const ScanClustering::Params& scanClusteringParams,
                         const ChainApproximation::Params& chainApproximationParams);

        std::vector<Features> onData(const armem::laser_scans::LaserScanStamped& data);

    public:
        static std::optional<VirtualRobot::MathTools::ConvexHull2D>
        convexHull(const Points& points);
        static std::optional<Ellipsoid>
        ellipsoid(const std::optional<VirtualRobot::MathTools::ConvexHull2D>& hull);
        static std::optional<Circle> circle(const Points& circle);
        static std::optional<Points>
        chainApproximation(const Points& points,
                           const std::optional<VirtualRobot::MathTools::ConvexHull2D>& convexHull,
                           const ChainApproximation::Params& params);


    private:
        std::vector<Features> features(const LaserScan& scan) const;

        std::vector<LaserScan> detectClusters(const LaserScan& scan) const;

        const ScanClustering::Params scanClusteringParams;
        const ChainApproximation::Params chainApproximationParams;
    };
} // namespace armarx::navigation::components::laser_scanner_feature_extraction
