/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::NavigationMemory
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <cstdint>
#include <mutex>
#include <string>

#include <Ice/Current.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/interface/core/PackagePath.h>

#include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>
#include <ArmarXGui/libraries/RemoteGui/Client/Tab.h>
#include <ArmarXGui/libraries/RemoteGui/Client/Widgets.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectPoseClient.h>
#include <RobotAPI/libraries/ArmarXObjects/plugins/ObjectPoseClientPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/ArVizComponentPlugin.h>
#include <RobotAPI/libraries/armem/server/plugins.h>
#include <RobotAPI/libraries/armem/server/plugins/ReadWritePluginUser.h>

#include <armarx/navigation/components/navigation_memory/ComponentInterface.h>

namespace armarx::navigation::components::navigation_memory
{

    /**
     * @defgroup Component-NavigationMemory NavigationMemory
     * @ingroup armarx_navigation-Components
     * A description of the component NavigationMemory.
     *
     * @class NavigationMemory
     * @ingroup Component-NavigationMemory
     * @brief Brief description of class NavigationMemory.
     *
     * Detailed description of class NavigationMemory.
     */
    class Component :
        virtual public armarx::Component,
        // , virtual public armarx::DebugObserverComponentPluginUser
        virtual public armarx::LightweightRemoteGuiComponentPluginUser,
        virtual public armarx::ArVizComponentPluginUser,
        virtual public armarx::armem::server::ReadWritePluginUser,
        virtual public memory::NavigationMemoryInterface,
        virtual public ObjectPoseClientPluginUser
    {
    public:
        Component();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        static std::string GetDefaultName();
        bool storeLocationGraph(const armarx::data::PackagePath& packagePath,
                                const Ice::Current& current) override;


    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;


        void createRemoteGuiTab();
        void RemoteGui_update() override;


    private:
        void visuRun();
        void loadSnapshot();


    private:
        struct CoreSegmentProperties
        {
            std::int64_t maxHistorySize = 100;
        };

        /// Properties shown in the Scenario GUI.
        struct Properties
        {
            std::string packageToLoad = "PriorKnowledgeData";
            std::string snapshotToLoad = "";

            struct LocationGraph
            {
                bool visuLocations = true;
                bool visuGraphEdges = true;
                bool visuRelativeLocations = true;
            };

            float visuFrequency = 2;

            bool visuCostmaps = true;
            float zOffsetCostmap = 5; // [mm]

            bool visuHumans = true;
            bool visuTransparent = false;
            int visuHumanMaxAgeMs = 1000;

            bool visuRooms = true;

            LocationGraph locationGraph;

            struct CoreSegments
            {
                CoreSegmentProperties parameterization;
                CoreSegmentProperties costmap{.maxHistorySize = 20};
                CoreSegmentProperties resultsGlobalPlanner;
                CoreSegmentProperties resultsLocalPlanner;
                CoreSegmentProperties events;
                CoreSegmentProperties exceptions;
                CoreSegmentProperties location;
                CoreSegmentProperties graph;
                CoreSegmentProperties human;
                CoreSegmentProperties laserScannerFeatures{.maxHistorySize = 20};
                CoreSegmentProperties rooms;

            } coreSeg;
        };

        Properties properties;
        mutable std::mutex propertiesMutex;

        /// Tab shown in the Remote GUI.
        struct RemoteGuiTab : armarx::RemoteGui::Client::Tab
        {
            struct LocationGraph
            {
                armarx::RemoteGui::Client::GroupBox group;

                armarx::RemoteGui::Client::Button reloadSnapshot;
                armarx::RemoteGui::Client::CheckBox visuLocations;
                armarx::RemoteGui::Client::CheckBox visuGraphEdges;

                void setup(Component& owner);
                void update(Component& owner) const;
            };

            LocationGraph locationGraph;
        };

        RemoteGuiTab tab;

        struct Tasks
        {
            SimpleRunningTask<>::pointer_type visuTask;
        };

        Tasks tasks;

        objpose::ObjectPoseClient objClient;
    };

} // namespace armarx::navigation::components::navigation_memory
