/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::ArmarXObjects::navigation_skill_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "Component.h"

#include <experimental/memory>
#include <mutex>
#include <string>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include <ArmarXGui/libraries/RemoteGui/Client/Widgets.h>

#include <RobotAPI/libraries/skills/core/SkillID.h>

#include <armarx/navigation/safety_guard/LaserBasedProximity.h>
#include <armarx/navigation/skills/GuideHumanToRoom.h>
#include <armarx/navigation/skills/MoveRelativePlanar.h>
#include <armarx/navigation/skills/MoveXMeters.h>
#include <armarx/navigation/skills/NavigateTo.h>
#include <armarx/navigation/skills/NavigateToChargingStation.h>
#include <armarx/navigation/skills/NavigateToLocation.h>
#include <armarx/navigation/skills/NavigateToNamedLocation.h>
#include <armarx/navigation/skills/RotateXDegrees.h>
#include <armarx/navigation/skills/constants.h>

namespace armarx::navigation::components::navigation_skill_provider
{
    Component::Component()
    {
        addPlugin(virtualRobotReaderPlugin);
        addPlugin(costmapReaderPlugin);
        addPlugin(graphReaderPlugin);
        addPlugin(roomsReaderPlugin);
    }

    const std::string Component::defaultName = skills::constants::NavigationSkillProviderName;

    namespace
    {
        void
        defineProximityFieldParams(const armarx::PropertyDefinitionsPtr& def,
                                   const std::string& prefix,
                                   safety_guard::ProximityFieldParams& params)
        {
            def->optional(params.reduceVelocity,
                          prefix + ".reduceVelocity",
                          "Whether to first reduce the velocity when coming closer to an obstacle "
                          "before stopping.");
            def->optional(params.minDistance,
                          prefix + ".minDistance",
                          "The minimum distance, below which the robot will stop.");
            def->optional(
                params.maxDistance,
                prefix + ".maxDistance",
                "The maximum distance, above which the robot will move with the maximum velocity.");
            def->optional(
                   params.k,
                   prefix + ".k",
                   "A parameter to adjust how the velocity is reduced when coming closer to "
                   "an obstacle. A higher value means the robot will reduce it's speed later.")
                .setMin(1);
            def->optional(
                   params.lambda,
                   prefix + ".lambda",
                   "A parameter to adjust how the velocity is reduced when coming closer to an "
                   "obstacle. A higher value means the robot will reduce it's speed earlier.")
                .setMin(1);
        }
    } // namespace

    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());

        def->component(navigatorPrx, "navigator");

        def->required(properties.robotName, "RobotName", "Default robot name.");

        auto optionalSubSkillID =
            [&def](armarx::skills::SkillID& skillID, const std::string& nameBase)
        {
            ARMARX_CHECK(skillID.providerId.has_value());
            def->optional(skillID.providerId->providerName, nameBase + ".providerName");
            def->optional(skillID.skillName, nameBase + ".skillName");
        };

        def->required(properties.safetyGuardParams.robotRadius,
                      "p.navigateToLocation.safetyGuardParams.robotRadius",
                      "The robot radius used for distance calculation to the platform.");
        def->optional(properties.safetyGuardParams.enableHumans,
                      "p.navigateToLocation.safetyGuardParams.enableHumans",
                      "Whether to consider humans for the safety guard.");
        def->optional(properties.safetyGuardParams.enableLaserScanners,
                      "p.navigateToLocation.safetyGuardParams.enableLaserscanners",
                      "Whether to consider laser scanners for the safety guard.");
        def->optional(properties.safetyGuardParams.maxVelocity.linear,
                      "p.navigateToLocation.safetyGuardParams.maxVelocityLinear",
                      "The maximum linear velocity.");
        def->optional(properties.safetyGuardParams.maxVelocity.angular,
                      "p.navigateToLocation.safetyGuardParams.maxVelocityAngular",
                      "The maximum angular velocity.");
        defineProximityFieldParams(def,
                                   "p.navigateToLocation.safetyGuardParams.humanProximityField",
                                   properties.safetyGuardParams.humanProximityField);
        defineProximityFieldParams(
            def,
            "p.navigateToLocation.safetyGuardParams.laserScannerProximityField",
            properties.safetyGuardParams.laserScannerProximityField);


        optionalSubSkillID(properties.navigateToNamedLocation.subSkillIDs.navigateToLocation,
                           "p.navigateToNamedLocation.subSkillIDs.navigateToLocation");
        optionalSubSkillID(properties.navigateToChargingStation.subSkillIDs.navigateToNamedLocation,
                           "p.navigateToChargingStation.subSkillIDs.navigateToNamedLocation");

        return def;
    }

    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requies the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);
    }

    void
    Component::onConnectComponent()
    {
        // Do things after connecting to topics and components.

        iceNavigator.setNavigatorComponent(navigatorPrx);

        {
            skills::NavigateTo::Services srv{
                .iceNavigator = iceNavigator,
                .memoryNameSystem = memoryNameSystem(),
            };
            addSkillFactory<skills::NavigateTo>(srv);
        }

        {
            skills::NavigateToLocation::Properties prop{
                .safetyGuardParamsMutex = std::experimental::make_observer(&safetyGuardParamsMutex),
                .safetyGuardParams =
                    std::experimental::make_observer(&properties.safetyGuardParams)};
            skills::NavigateToLocation::Services srv{
                .iceNavigator = iceNavigator,
                .memoryNameSystem = memoryNameSystem(),
            };
            addSkillFactory<skills::NavigateToLocation>(prop, srv);
        }

        {
            skills::NavigateToNamedLocation::Services services{
                .locationReader = graphReaderPlugin->get(),
            };
            addSkillFactory<skills::NavigateToNamedLocation>(properties.navigateToNamedLocation,
                                                             services);
        }

        {
            addSkillFactory<skills::NavigateToChargingStation>(
                properties.navigateToChargingStation);
        }

        {
            skills::MoveXMeters::Services srv{
                .iceNavigator = iceNavigator,
                .memoryNameSystem = memoryNameSystem(),
            };
            addSkillFactory<skills::MoveXMeters>(srv);
        }

        {

            skills::RotateXDegrees::Services srv{
                .iceNavigator = iceNavigator,
                .memoryNameSystem = memoryNameSystem(),
            };
            addSkillFactory<skills::RotateXDegrees>(srv);
        }

        {
            skills::MoveRelativePlanar::Properties properties{
                .robotName = this->properties.robotName,
            };
            skills::MoveRelativePlanar::Services srv{
                .iceNavigator = iceNavigator,
                .memoryNameSystem = memoryNameSystem(),
                .robotReader = virtualRobotReaderPlugin->get(),
            };
            addSkillFactory<skills::MoveRelativePlanar>(properties, srv);
        }

        {
            ARMARX_CHECK_NOT_NULL(virtualRobotReaderPlugin);
            ARMARX_CHECK_NOT_NULL(virtualRobotReaderPlugin);
            ARMARX_CHECK_NOT_NULL(roomsReaderPlugin);

            skills::GuideHumanToRoom::Services srv{.iceNavigator = iceNavigator,
                                                   .memoryNameSystem = memoryNameSystem(),
                                                   .virtualRobotReader =
                                                       virtualRobotReaderPlugin->get(),
                                                   .costmapReader = costmapReaderPlugin->get(),
                                                   .roomsReader = roomsReaderPlugin->get(),
                                                   .arviz = arviz};

            addSkillFactory<skills::GuideHumanToRoom>(srv);
        }


        /* (Requies the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        /* (Requires the armarx::ArVizComponentPluginUser.)
        // Draw boxes in ArViz.
        // (Before starting any threads, we don't need to lock mutexes.)
        drawBoxes(properties, arviz);
        */

        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }
    }

    void
    Component::onDisconnectComponent()
    {
    }

    void
    Component::onExitComponent()
    {
    }

    std::string
    Component::getDefaultName() const
    {
        return Component::defaultName;
    }

    std::string
    Component::GetDefaultName()
    {
        return Component::defaultName;
    }

    void
    Component::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        VBoxLayout root;

        // thread-safe access to params
        const auto safetyGuardParams = [this]()
        {
            std::lock_guard g{safetyGuardParamsMutex};
            return properties.safetyGuardParams;
        }();

        const auto setSpinBox = [](armarx::RemoteGui::Client::FloatSpinBox& w,
                                   float min,
                                   float max,
                                   int decimals,
                                   float step)
        {
            w.setRange(min, max);
            w.setDecimals(decimals);
            w.setSteps(static_cast<int>((max - min) / step));
        };

        {
            GridLayout grid;
            int row = 0;
            grid.add(Label("Enable Humans"), {.row = row, .column = 0})
                .add(tab.enableHumans, {.row = row++, .column = 1});
            grid.add(Label("Enable Laserscanner"), {.row = row, .column = 0})
                .add(tab.enableLaserScanners, {.row = row++, .column = 1});
            grid.add(armarx::RemoteGui::Client::VSpacer(), {.row = row++, .column = 0});
            grid.add(Label("Max velocity linear"), {.row = row, .column = 0})
                .add(tab.maxVelocityLinear, {.row = row++, .column = 1});
            grid.add(Label("Max velocity angular"), {.row = row, .column = 0})
                .add(tab.maxVelocityAngular, {.row = row++, .column = 1});
            root.addChild(grid);

            tab.enableHumans.setValue(safetyGuardParams.enableHumans);
            tab.enableLaserScanners.setValue(safetyGuardParams.enableLaserScanners);
            tab.maxVelocityLinear.setValue(safetyGuardParams.maxVelocity.linear);
            setSpinBox(tab.maxVelocityLinear, 0, 2000, 1, 1);
            tab.maxVelocityAngular.setValue(safetyGuardParams.maxVelocity.angular);
            setSpinBox(tab.maxVelocityAngular, 0, 5, 2, 0.1);
        }
        root.addChild(VSpacer());

        const auto setupProximityFieldGui =
            [&root, &setSpinBox](RemoteGuiTab::ProximityField& field,
                                 const safety_guard::ProximityFieldParams& values,
                                 const std::string& label)
        {
            GridLayout grid;
            int row = 0;
            grid.add(Label("Reduce speed"), {.row = row, .column = 0})
                .add(field.reduceSpeed, {.row = row++, .column = 1});
            grid.add(Label("Min distance"), {.row = row, .column = 0})
                .add(field.minDistance, {.row = row++, .column = 1});
            grid.add(armarx::RemoteGui::Client::VSpacer(), {.row = row++, .column = 0});
            grid.add(Label("Max distance"), {.row = row, .column = 0})
                .add(field.maxDistance, {.row = row++, .column = 1});
            grid.add(Label("k"), {.row = row, .column = 0})
                .add(field.k, {.row = row++, .column = 1});
            grid.add(Label("lambda"), {.row = row, .column = 0})
                .add(field.lambda, {.row = row++, .column = 1});

            GroupBox box({grid});
            box.setLabel(label);
            root.addChild(box);

            field.reduceSpeed.setValue(values.reduceVelocity);
            field.minDistance.setValue(values.minDistance);
            setSpinBox(field.minDistance, 0, 5000, 1, 1);
            field.maxDistance.setValue(values.maxDistance);
            setSpinBox(field.maxDistance, 0, 10000, 1, 1);
            field.k.setValue(values.k);
            setSpinBox(field.k, 1, 20, 2, 0.1);
            field.lambda.setValue(values.lambda);
            setSpinBox(field.lambda, 1, 20, 2, 0.1);
        };

        setupProximityFieldGui(tab.humanProximityField,
                               safetyGuardParams.humanProximityField,
                               "Human proximity field");
        root.addChild(VSpacer());
        setupProximityFieldGui(tab.laserScannerProximityField,
                               safetyGuardParams.laserScannerProximityField,
                               "LaserScanner proximity field");

        RemoteGui_createTab(getName(), root, &tab);
    }

    void
    Component::RemoteGui_update()
    {
        std::lock_guard g{safetyGuardParamsMutex};

        auto& safetyGuardParams = properties.safetyGuardParams;

        safetyGuardParams.enableHumans = tab.enableHumans.getValue();
        safetyGuardParams.enableLaserScanners = tab.enableLaserScanners.getValue();
        safetyGuardParams.maxVelocity.linear = tab.maxVelocityLinear.getValue();
        safetyGuardParams.maxVelocity.angular = tab.maxVelocityAngular.getValue();

        const auto readProximityField = [&](const RemoteGuiTab::ProximityField& field,
                                            safety_guard::ProximityFieldParams& values)
        {
            values.reduceVelocity = field.reduceSpeed.getValue();
            values.minDistance = field.minDistance.getValue();
            values.maxDistance = field.maxDistance.getValue();
            values.k = field.k.getValue();
            values.lambda = field.lambda.getValue();
        };

        readProximityField(tab.humanProximityField, safetyGuardParams.humanProximityField);
        readProximityField(tab.laserScannerProximityField,
                           safetyGuardParams.laserScannerProximityField);
    }

    /* (Requires the armarx::ArVizComponentPluginUser.)
    void
    Component::drawBoxes(const Component::Properties& p, viz::Client& arviz)
    {
        // Draw something in ArViz (requires the armarx::ArVizComponentPluginUser.
        // See the ArVizExample in RobotAPI for more examples.

        viz::Layer layer = arviz.layer(p.boxLayerName);
        for (int i = 0; i < p.numBoxes; ++i)
        {
            layer.add(viz::Box("box_" + std::to_string(i))
                      .position(Eigen::Vector3f(i * 100, 0, 0))
                      .size(20).color(simox::Color::blue()));
        }
        arviz.commit(layer);
    }
    */


    ARMARX_REGISTER_COMPONENT_EXECUTABLE(Component, Component::GetDefaultName());

} // namespace armarx::navigation::components::navigation_skill_provider
