/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::ArmarXObjects::navigation_skill_provider
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once

#include <mutex>
#include <string>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>

#include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>
#include <ArmarXGui/libraries/RemoteGui/Client/Tab.h>
#include <ArmarXGui/libraries/RemoteGui/Client/Widgets.h>

#include <RobotAPI/libraries/RobotAPIComponentPlugins/ArVizComponentPlugin.h>
#include <RobotAPI/libraries/armem/client/forward_declarations.h>
#include <RobotAPI/libraries/armem/client/plugins/ListeningPluginUser.h>
#include <RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>
#include <RobotAPI/libraries/skills/provider/SkillProviderComponentPlugin.h>

#include <armarx/navigation/client/ice/NavigatorInterface.h>
#include <armarx/navigation/client/services/IceNavigator.h>
#include <armarx/navigation/components/navigation_skill_provider/ComponentInterface.h>
#include <armarx/navigation/memory/client/costmap/Reader.h>
#include <armarx/navigation/memory/client/graph/Reader.h>
#include <armarx/navigation/memory/client/rooms/Reader.h>
#include <armarx/navigation/safety_guard/LaserBasedProximity.h>
#include <armarx/navigation/skills/NavigateToChargingStation.h>
#include <armarx/navigation/skills/NavigateToNamedLocation.h>
#include <armarx/navigation/skills/forward_declarations.h>

namespace armarx::navigation::components::navigation_skill_provider
{

    class Component :
        virtual public armarx::Component,
        virtual public ComponentInterface,
        virtual public armarx::SkillProviderComponentPluginUser,
        // virtual public armarx::navigation::client::ComponentPluginUser,
        virtual public armem::ListeningClientPluginUser,
        // , virtual public armarx::DebugObserverComponentPluginUser
        virtual public armarx::LightweightRemoteGuiComponentPluginUser,
        virtual public armarx::ArVizComponentPluginUser
    {
    public:
        Component();
        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        /// Get the component's default name.
        static std::string GetDefaultName();


    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;


        /// This function should be called once in onConnect() or when you
        /// need to re-create the Remote GUI tab.
        void createRemoteGuiTab();

        /// After calling `RemoteGui_startRunningTask`, this function is
        /// called periodically in a separate thread. If you update variables,
        /// make sure to synchronize access to them.
        void RemoteGui_update() override;


    private:
        // Private methods go here.

        // Forward declare `Properties` if you used it before its defined.
        // struct Properties;

        /* (Requires the armarx::ArVizComponentPluginUser.)
        /// Draw some boxes in ArViz.
        void drawBoxes(const Properties& p, viz::Client& arviz);
        */


    private:
        static const std::string defaultName;

        /// Properties shown in the Scenario GUI.
        struct Properties
        {
            std::string robotName;

            safety_guard::LaserBasedProximityParams safetyGuardParams;

            skills::NavigateToNamedLocation::Properties navigateToNamedLocation;

            skills::NavigateToChargingStation::Properties navigateToChargingStation;
        };

        std::mutex safetyGuardParamsMutex;

        Properties properties;


        client::IceNavigator iceNavigator;

        client::NavigatorInterfacePrx navigatorPrx;

        template <typename T>
        using ReaderWriterPlugin = armem::client::plugins::ReaderWriterPlugin<T>;


        ReaderWriterPlugin<armem::robot_state::VirtualRobotReader>* virtualRobotReaderPlugin =
            nullptr;
        ReaderWriterPlugin<memory::client::costmap::Reader>* costmapReaderPlugin = nullptr;
        ReaderWriterPlugin<memory::client::graph::Reader>* graphReaderPlugin = nullptr;
        ReaderWriterPlugin<memory::client::rooms::Reader>* roomsReaderPlugin = nullptr;

        struct RemoteGuiTab : armarx::RemoteGui::Client::Tab
        {
            struct ProximityField
            {
                armarx::RemoteGui::Client::CheckBox reduceSpeed;
                armarx::RemoteGui::Client::FloatSpinBox minDistance;
                armarx::RemoteGui::Client::FloatSpinBox maxDistance;
                armarx::RemoteGui::Client::FloatSpinBox k;
                armarx::RemoteGui::Client::FloatSpinBox lambda;
            };

            armarx::RemoteGui::Client::CheckBox enableHumans;
            armarx::RemoteGui::Client::CheckBox enableLaserScanners;

            armarx::RemoteGui::Client::FloatSpinBox maxVelocityLinear;
            armarx::RemoteGui::Client::FloatSpinBox maxVelocityAngular;

            ProximityField humanProximityField;
            ProximityField laserScannerProximityField;
        };

        RemoteGuiTab tab;
    };

} // namespace armarx::navigation::components::navigation_skill_provider
