/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Tobias Gröger ( tobias dot groeger at student dot kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <optional>
#include <string>
#include <vector>

#include <RobotAPI/libraries/ArmarXObjects/ObjectInfo.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectPoseClient.h>
#include <RobotAPI/libraries/ArmarXObjects/forward_declarations.h>
#include <RobotAPI/libraries/core/FramedPose.h>

#include <armarx/navigation/core/basic_types.h>

namespace armarx::navigation::core
{

    struct ResolvedLocation
    {
        std::optional<Pose> pose;
        std::string errorMsg; // error message in case the pose is not set

        ResolvedLocation(const Pose& pose) : pose(pose)
        {
        }

        ResolvedLocation(const std::string& errorMsg) : errorMsg(errorMsg)
        {
        }
    };

    /*!
     * \brief Resolve the FramedPose of a location to a global pose.
     * \param instanceID instance to use if the location is relative to an object class
     *
     * When the location is relative to a specific instance the parameter instanceID is ignored.
     * Otherwise (the location is relative to a object class) it is set as the instance of that class.
     * The object pose and info is fetched through a network call.
     * For a cached version see below.
     */
    ResolvedLocation resolveLocation(const objpose::ObjectPoseClient& client,
                                     const FramedPose& framedPose,
                                     const std::string& instanceID = "");

    /*!
     * \brief Resolve the FramedPose of a location to a global pose.
     * \param instanceID instance to use if the location is relative to an object class
     *
     * When the location is relative to a specific instance the parameter instanceID is ignored.
     * Otherwise (the location is relative to a object class) it is set as the instance of that class.
     */
    ResolvedLocation resolveLocation(const objpose::ObjectPoseMap& objects,
                                     const std::vector<ObjectInfo>& info,
                                     const FramedPose& framedPose,
                                     const std::string& instanceID = "");

} // namespace armarx::navigation::core
