/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::global_planning
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <SimoxUtility/math/convert/mat4f_to_xyyaw.h>

#include <ArmarXCore/core/logging/Logging.h>

#include <armarx/navigation/conversions/eigen.h>

#include <range/v3/range/conversion.hpp>
#include <range/v3/view/transform.hpp>
#define BOOST_TEST_MODULE Navigation::ArmarXLibraries::global_planning

#define ARMARX_BOOST_TEST

#include <cmath>
#include <cstddef>
#include <iostream>

#include <boost/test/tools/old/interface.hpp>

#include <Eigen/Core>
#include <Eigen/Geometry>

#include <SimoxUtility/math/convert/mat3f_to_rpy.h>
#include <VirtualRobot/MathTools.h>

#include <ArmarXCore/core/exceptions/LocalException.h>

#include <armarx/navigation/Test.h>
#include <armarx/navigation/core/basic_types.h>
#include <armarx/navigation/global_planning/optimization/OrientationOptimizer.h>

#include <sciplot/sciplot.hpp>

void
printStartOfTestcase(std::string name)
{
    ARMARX_INFO << "\n\n\n\nRunning test '" << name << "'\n";
}

double
toYaw(const armarx::navigation::core::GlobalTrajectoryPoint& pt)
{
    return simox::math::mat4f_to_xyyaw(pt.waypoint.pose.matrix()).z();
}

double
absoluteMinimalRotationDifference(double r1, double r2)
{
    return std::abs(VirtualRobot::MathTools::AngleDelta(r1, r2));
    //  ARMARX_INFO << "r1: " << r1;
    //  const double r1_normalized = std::fmod(r1, M_2_PI);
    //  ARMARX_INFO << "r1 normalized: " << r1_normalized;
    //  const double r2_normalized = std::fmod(r2, M_2_PI);
    //  const double deltaOneWay = std::fmod(r1_normalized - r2_normalized, M_2_PI);
    //  const double deltaOtherWay = std::fmod(r2_normalized - r1_normalized, M_2_PI);
    //  return std::min(deltaOneWay, deltaOtherWay);
}

double
absoluteTotalRotation(const armarx::navigation::core::GlobalTrajectory& trajectory)
{
    const std::vector<double> orientations =
        trajectory.points() | ranges::views::transform(toYaw) | ranges::to_vector;

    double absoluteTotalRotation = 0.f;

    BOOST_REQUIRE(not orientations.empty());
    for (std::size_t i = 0; i < orientations.size() - 1; i++)
    {
        absoluteTotalRotation +=
            absoluteMinimalRotationDifference(orientations.at(i), orientations.at(i + 1));
    }
    return absoluteTotalRotation;
}

bool
moreTurningThanNecessary(const armarx::navigation::core::GlobalTrajectory& trajectory,
                         double toleranceFactor = 1.5f)
{
    BOOST_REQUIRE(trajectory.points().size() > 1);
    double initialOrientation = toYaw(trajectory.points().front());
    double finalOrientation = toYaw(trajectory.points().back());
    double directDelta = absoluteMinimalRotationDifference(initialOrientation, finalOrientation);

    ARMARX_INFO << "Direct rotational difference: " << directDelta;
    ARMARX_INFO << "Total rotation via trajectory: " << absoluteTotalRotation(trajectory);

    return absoluteTotalRotation(trajectory) > directDelta * toleranceFactor;
}

void
plotResult(const armarx::navigation::core::GlobalTrajectory& trajectory,
           const std::vector<double>& orientations,
           const std::vector<double>& priors)
{
    sciplot::Plot2D plot;

    plot.xlabel("x");
    plot.ylabel("y");
    plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);

    std::vector<float> waypoint_x, waypoint_y;
    const auto poses = trajectory.poses();
    for (std::size_t i = 0; i < poses.size(); i++)
    {
        const auto angleToVec = [](double angle)
        {
            angle += M_PI_2;
            return Eigen::Vector2f{std::cos(angle), std::sin(angle)};
        };

        const armarx::navigation::core::Pose2D pose = armarx::navigation::conv::to2D(poses.at(i));
        const Eigen::Vector2f waypoint = pose.translation();

        waypoint_x.push_back(waypoint.x());
        waypoint_y.push_back(waypoint.y());

        const auto drawArrow =
            [](const Eigen::Vector2f& from, const Eigen::Vector2f& to, const std::string& color)
        {
            const auto formatLocation = [](const Eigen::Vector2f& loc)
            { return std::to_string(loc.x()) + "," + std::to_string(loc.y()); };

            return "set arrow from " + formatLocation(from) + " to " + formatLocation(to) +
                   +"head lw 1 lc rgb \"" + color + "\"";
        };

        const float arrowLength = 500;

        double theta = Eigen::Rotation2Df(pose.linear()).angle();
        Eigen::Vector2f arrowT = angleToVec(theta) * arrowLength;
        arrowT += waypoint;
        plot.gnuplot(drawArrow(waypoint, arrowT, "blue"));

        double initial = orientations.at(i);
        Eigen::Vector2f initialArrowT = angleToVec(initial) * (4.0 / 5.0 * arrowLength);
        initialArrowT += waypoint;
        plot.gnuplot(drawArrow(waypoint, initialArrowT, "red"));

        double prior = priors.at(i);
        Eigen::Vector2f priorArrowT = angleToVec(prior) * (3.0 / 5.0 * arrowLength);
        priorArrowT += waypoint;
        plot.gnuplot(drawArrow(waypoint, priorArrowT, "green"));
    }

    plot.fontSize(12);
    plot.drawCurveWithPoints(waypoint_x, waypoint_y).label("Result").lineColor("blue");
    plot.drawCurve(std::vector<float>{0.}, std::vector<float>{0.})
        .label("Initial")
        .lineColor("red");
    plot.drawCurve(std::vector<float>{0.}, std::vector<float>{0.})
        .label("Prior")
        .lineColor("green");
    plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2).frameLineStyle(1);
    plot.gnuplot("set autoscale xfix; set autoscale yfix; set offsets 0, 500, 500, 0");
    plot.gnuplot("set size ratio -1");


    sciplot::Figure fig = {{plot}};
    // Create canvas to hold figure
    sciplot::Canvas canvas = {{fig}};
    canvas.size(1000, 1000);

    // Show the plot in a pop-up window
    canvas.show();
}

void
plotResult(armarx::navigation::global_planning::OrientationOptimizer::OptimizationResult result)
{
    /* //COMMENT IN FOR TEST CASE PLOTTING
     * //YOU NEED TO COMMENT IN 1 SECTION IN OrientationOptimizer.h AND 3 SECTIONS IN OrientationOptimizer.cpp
    plotResult(result.trajectory.value(), result.initial, result.prior); */
}

//==================================================================================================
//----------------------------- START OF ALONG-PATH-TESTCASES --------------------------------------
//==================================================================================================

BOOST_AUTO_TEST_CASE(testOrientationOptimizationYDecreasing, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationYDecreasing");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;


    const Eigen::Isometry3f start(Eigen::Translation3f{5000, 350, 0} *
                                  Eigen::AngleAxis(-1.4f, Eigen::Vector3f::UnitZ()));

    std::vector<nav::core::Position> waypoints;
    waypoints.emplace_back(Eigen::Vector3f{4800, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{4600, 250, 0});
    waypoints.emplace_back(Eigen::Vector3f{4400, 150, 0});
    waypoints.emplace_back(Eigen::Vector3f{4200, 50, 0});
    waypoints.emplace_back(Eigen::Vector3f{4000, -50, 0});

    waypoints.emplace_back(Eigen::Vector3f{3800, -150, 0});
    waypoints.emplace_back(Eigen::Vector3f{3600, -250, 0});
    waypoints.emplace_back(Eigen::Vector3f{3400, -350, 0});
    waypoints.emplace_back(Eigen::Vector3f{3200, -450, 0});
    waypoints.emplace_back(Eigen::Vector3f{3000, -550, 0});

    waypoints.emplace_back(Eigen::Vector3f{2800, -650, 0});
    waypoints.emplace_back(Eigen::Vector3f{2600, -750, 0});
    waypoints.emplace_back(Eigen::Vector3f{2400, -850, 0});
    waypoints.emplace_back(Eigen::Vector3f{2200, -950, 0});

    const Eigen::Isometry3f goal(Eigen::Translation3f{2000.00, -950.00, 0} *
                                 Eigen::AngleAxis(1.4f, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(start, waypoints, goal, 500);


    gp::OrientationOptimizer::Params params;
    params.predefinedRotationDirection = gp::RotationDirection::Clockwise;

    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
    plotResult(result);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationYConstant, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationYConstant");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    const Eigen::Isometry3f start(Eigen::Translation3f{5000, 350, 0} *
                                  Eigen::AngleAxis(-1.4f, Eigen::Vector3f::UnitZ()));

    std::vector<nav::core::Position> waypoints;
    waypoints.emplace_back(Eigen::Vector3f{4800, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{4600, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{4400, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{4200, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{4000, 350, 0});

    waypoints.emplace_back(Eigen::Vector3f{3800, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{3600, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{3400, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{3200, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{3000, 350, 0});

    waypoints.emplace_back(Eigen::Vector3f{2800, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{2600, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{2400, 350, 0});
    waypoints.emplace_back(Eigen::Vector3f{2200, 350, 0});

    const Eigen::Isometry3f goal(Eigen::Translation3f{2000.00, 350.00, 0} *
                                 Eigen::AngleAxis(1.4f, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(start, waypoints, goal, 500);

    gp::OrientationOptimizer::Params params;
    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
    plotResult(result);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationYIncreasing, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationYIncreasing");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;


    const Eigen::Isometry3f start(Eigen::Translation3f{5000, 350, 0} *
                                  Eigen::AngleAxis(-1.4f, Eigen::Vector3f::UnitZ()));

    std::vector<nav::core::Position> waypoints;
    waypoints.emplace_back(Eigen::Vector3f{4800, 450, 0});
    waypoints.emplace_back(Eigen::Vector3f{4600, 550, 0});
    waypoints.emplace_back(Eigen::Vector3f{4400, 650, 0});
    waypoints.emplace_back(Eigen::Vector3f{4200, 750, 0});
    waypoints.emplace_back(Eigen::Vector3f{4000, 850, 0});

    waypoints.emplace_back(Eigen::Vector3f{3800, 950, 0});
    waypoints.emplace_back(Eigen::Vector3f{3600, 1050, 0});
    waypoints.emplace_back(Eigen::Vector3f{3400, 1150, 0});
    waypoints.emplace_back(Eigen::Vector3f{3200, 1250, 0});
    waypoints.emplace_back(Eigen::Vector3f{3000, 1350, 0});

    waypoints.emplace_back(Eigen::Vector3f{2800, 1450, 0});
    waypoints.emplace_back(Eigen::Vector3f{2600, 1550, 0});
    waypoints.emplace_back(Eigen::Vector3f{2400, 1650, 0});
    waypoints.emplace_back(Eigen::Vector3f{2200, 1750, 0});

    const Eigen::Isometry3f goal(Eigen::Translation3f{2000.00, 1850.00, 0} *
                                 Eigen::AngleAxis(1.4f, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(start, waypoints, goal, 500);

    gp::OrientationOptimizer::Params params;
    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
    plotResult(result);
}

/*
 * This test case is the result of a trajectory received upon navigating from the R003 workbench to
 * the mobile kitchen. It was found via trial and error as a path that results in faulty orientation
 * optimization (rotating in the wrong direction first, then very quickly rotating in the other
 * direction).
 * Note: Appearing bug got (hopefully) fixed
 */
BOOST_AUTO_TEST_CASE(testOrientationOptimizationEmpiricallyObservedFailure,
                     *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationEmpiricallyObservedFailure");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    // start pose (workbench)
    const Eigen::Isometry3f start(Eigen::Translation3f{4933.27, 357.81, 0} *
                                  Eigen::AngleAxis(-1.60762f, Eigen::Vector3f::UnitZ()));

    std::vector<nav::core::Position> waypoints;
    waypoints.emplace_back(Eigen::Vector3f{4752.28, 442.90, 0});
    waypoints.emplace_back(Eigen::Vector3f{4608.42, 581.85, 0});
    waypoints.emplace_back(Eigen::Vector3f{4466.01, 722.27, 0});
    waypoints.emplace_back(Eigen::Vector3f{4324.47, 863.57, 0});

    waypoints.emplace_back(Eigen::Vector3f{4182.99, 1004.94, 0});
    waypoints.emplace_back(Eigen::Vector3f{4041.55, 1146.34, 0});
    waypoints.emplace_back(Eigen::Vector3f{3863.20, 1236.84, 0});
    waypoints.emplace_back(Eigen::Vector3f{3717.90, 1374.28, 0});
    waypoints.emplace_back(Eigen::Vector3f{3574.78, 1513.98, 0});

    waypoints.emplace_back(Eigen::Vector3f{3433.10, 1655.14, 0});
    waypoints.emplace_back(Eigen::Vector3f{3291.55, 1796.44, 0});
    waypoints.emplace_back(Eigen::Vector3f{3236.05, 1849.79, 0});
    waypoints.emplace_back(Eigen::Vector3f{3038.97, 1883.87, 0});
    waypoints.emplace_back(Eigen::Vector3f{2850.00, 1949.36, 0});

    waypoints.emplace_back(Eigen::Vector3f{2650.00, 1949.36, 0});
    waypoints.emplace_back(Eigen::Vector3f{2450.00, 1949.36, 0});
    waypoints.emplace_back(Eigen::Vector3f{2250.00, 1949.36, 0});
    waypoints.emplace_back(Eigen::Vector3f{2050.00, 1949.37, 0});
    waypoints.emplace_back(Eigen::Vector3f{1850.00, 1949.37, 0});

    waypoints.emplace_back(Eigen::Vector3f{1650.00, 1949.38, 0});
    waypoints.emplace_back(Eigen::Vector3f{1450.00, 1949.42, 0});
    waypoints.emplace_back(Eigen::Vector3f{1250.00, 1949.56, 0});

    // end pose (kitchen)
    const Eigen::Isometry3f goal(Eigen::Translation3f{1050.00, 1950.00, 0} *
                                 Eigen::AngleAxis(M_PI_2f32, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(start, waypoints, goal, 500);

    gp::OrientationOptimizer::Params params;
    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
    plotResult(result);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationEdgeCase /*, *boost::unit_test::disabled()*/)
{
    printStartOfTestcase("testOrientationOptimizationEdgeCase");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;


    const Eigen::Isometry3f start(Eigen::Translation3f{3000, 400, 0} *
                                  Eigen::AngleAxis(2.5f, Eigen::Vector3f::UnitZ()));

    std::vector<nav::core::Position> waypoints;
    waypoints.emplace_back(Eigen::Vector3f{2900, 400, 0});
    waypoints.emplace_back(Eigen::Vector3f{2800, 420, 0});
    waypoints.emplace_back(Eigen::Vector3f{2700, 470, 0});
    waypoints.emplace_back(Eigen::Vector3f{2650, 600, 0});
    waypoints.emplace_back(Eigen::Vector3f{2600, 900, 0});

    waypoints.emplace_back(Eigen::Vector3f{2600, 1100, 0});
    waypoints.emplace_back(Eigen::Vector3f{2600, 1400, 0});
    waypoints.emplace_back(Eigen::Vector3f{2550, 1550, 0});
    waypoints.emplace_back(Eigen::Vector3f{2500, 1650, 0});
    waypoints.emplace_back(Eigen::Vector3f{2400, 1700, 0});

    waypoints.emplace_back(Eigen::Vector3f{2200, 1750, 0});
    waypoints.emplace_back(Eigen::Vector3f{2000, 1800, 0});
    waypoints.emplace_back(Eigen::Vector3f{1800, 1800, 0});
    waypoints.emplace_back(Eigen::Vector3f{1600, 1800, 0});
    waypoints.emplace_back(Eigen::Vector3f{1500, 1750, 0});

    waypoints.emplace_back(Eigen::Vector3f{1400, 1650, 0});
    waypoints.emplace_back(Eigen::Vector3f{1350, 1500, 0});
    waypoints.emplace_back(Eigen::Vector3f{1300, 1300, 0});
    waypoints.emplace_back(Eigen::Vector3f{1300, 1100, 0});
    waypoints.emplace_back(Eigen::Vector3f{1300, 900, 0});

    waypoints.emplace_back(Eigen::Vector3f{1300, 700, 0});
    waypoints.emplace_back(Eigen::Vector3f{1300, 500, 0});

    const Eigen::Isometry3f goal(Eigen::Translation3f{1300, 300, 0} *
                                 Eigen::AngleAxis(1.6f, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(start, waypoints, goal, 500);

    gp::OrientationOptimizer::Params params;
    params.predefinedRotationDirection = gp::RotationDirection::CounterClockwise;

    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
    plotResult(result);
}

//==================================================================================================
//------------------------------ END OF ALONG-PATH-TESTCASES ---------------------------------------
//==================================================================================================


BOOST_AUTO_TEST_CASE(testOrientationOptimizationRotatedGoalMinimalPrior,
                     *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationRotatedGoalMinimalPrior");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1000});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 2000} *
                      Eigen::AngleAxis(M_PI_2f32, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);

    gp::OrientationOptimizer::Params params{.iterations = 1,
                                            .movementDirWeightStart = 0.0F,
                                            .movementDirWeightEnd = 1.F,
                                            .smoothnessWeight = 0.F,
                                            .smoothnessWeightStartGoal = 0.0F};


    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationRotatedGoalMinimalSmoothness,
                     *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationRotatedGoalMinimalSmoothness");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1000});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 2000} *
                      Eigen::AngleAxis(M_PI_2f32, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);

    gp::OrientationOptimizer::Params params{.iterations = 1,
                                            .movementDirWeightStart = 0.0F,
                                            .movementDirWeightEnd = 0.F,
                                            .smoothnessWeight = 1.F,
                                            .smoothnessWeightStartGoal = 0.0F};


    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationRotatedGoalMinimalSmoothnessStartGoal,
                     *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationRotatedGoalMinimalSmoothnessStartGoal");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1000});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 2000} *
                      Eigen::AngleAxis(M_PI_2f32, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);

    gp::OrientationOptimizer::Params params{.iterations = 1,
                                            .movementDirWeightStart = 0.0F,
                                            .movementDirWeightEnd = 0.F,
                                            .smoothnessWeight = 0.F,
                                            .smoothnessWeightStartGoal = 1.0F};


    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();
    BOOST_REQUIRE(result);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationRotatedGoal, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationRotatedGoal");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1000});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 1500});
    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 2000} *
                      Eigen::AngleAxis(M_PI_2f32, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);

    gp::OrientationOptimizer::Params params{};


    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();

    // # of points should be unchanged
    BOOST_REQUIRE_EQUAL(result.trajectory.value().poses().size(), trajectory.poses().size());

    // avoid an underflow in the for loop below
    BOOST_REQUIRE_GE(trajectory.poses().size(), 2);

    // loop over intermediate points
    for (std::size_t i = 1; i < (result.trajectory.value().poses().size() - 2); i++)
    {
        const auto rpy =
            simox::math::mat3f_to_rpy(result.trajectory.value().poses().at(i).linear());
        const auto yaw = rpy.z();

        // The orientation should be changed compared to the unoptimized version.
        // As the last pose is rotated by 90°, the intermediate poses should at least be
        // changed gentlty into this direction.
        BOOST_CHECK_GT(yaw, 0.1);
    }
}

BOOST_AUTO_TEST_CASE(testOrientationOptimizationCircular, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimizationCircular");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    for (int i = 0; i < 20; i++)
    {
        path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 500 * i} *
                          Eigen::AngleAxis(M_PI_2f32 * i, Eigen::Vector3f::UnitZ()));
    }

    path.emplace_back(Eigen::Translation3f{Eigen::Vector3f::UnitX() * 500 * 20});


    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);

    // Here we don't care about the orientatation prior.
    // The orientation of the intermediate points should converge to 0 (same direction as start and goal)
    gp::OrientationOptimizer::Params params{
        .iterations = 1, .movementDirWeightStart = 0.0F, .movementDirWeightEnd = 0.F};


    gp::OrientationOptimizer optimizer(trajectory, params);

    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();

    // # of points should be unchanged
    BOOST_REQUIRE_EQUAL(result.trajectory.value().poses().size(), trajectory.poses().size());

    // avoid an underflow in the for loop below
    BOOST_REQUIRE_GE(trajectory.poses().size(), 2);

    // loop over intermediate points
    for (std::size_t i = 0; i < (result.trajectory.value().poses().size() - 2); i++)
    {
        const auto rpy =
            simox::math::mat3f_to_rpy(result.trajectory.value().poses().at(i).linear());
        const auto yaw = rpy.z();

        // The orientation should be changed compared to the unoptimized version.
        // As the last pose is oriented the same way as the start, the intermediate poses should
        // have the same orientation. => yaw ~= 0
        BOOST_CHECK_LT(std::abs(yaw), 0.01);
    }
}

BOOST_AUTO_TEST_CASE(testOrientationOptimization2Points, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimization2Points");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    // The position should not matter

    constexpr float initialOrientation = 0;
    constexpr float finalOrientation = 2.18166;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    path.emplace_back(nav::core::Pose::Identity() *
                      Eigen::AngleAxisf(finalOrientation, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);


    gp::OrientationOptimizer::Params params;
    gp::OrientationOptimizer optimizer(trajectory, params);

    // only the middle point should be changed.
    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();

    const auto toYaw = [](const nav::core::GlobalTrajectoryPoint& pt) -> double
    { return simox::math::mat4f_to_xyyaw(pt.waypoint.pose.matrix()).z(); };

    const std::vector<double> orientations =
        result.trajectory->points() | ranges::views::transform(toYaw) | ranges::to_vector;

    ARMARX_INFO << VAROUT(orientations);

    // # of points should be unchanged
    BOOST_REQUIRE_CLOSE(orientations.front(), initialOrientation, 0.01);
    BOOST_REQUIRE_CLOSE(orientations.back(), finalOrientation, 0.01);
}

BOOST_AUTO_TEST_CASE(testOrientationOptimization3Points, *boost::unit_test::disabled())
{
    printStartOfTestcase("testOrientationOptimization3Points");
    namespace nav = armarx::navigation;
    namespace gp = armarx::navigation::global_planning;

    // The position should not matter

    constexpr float initialOrientation = 0;
    constexpr float finalOrientation = 2.18166;

    nav::core::Path path;
    path.emplace_back(nav::core::Pose::Identity());
    path.emplace_back(nav::core::Pose::Identity() *
                      Eigen::AngleAxisf(-1.65832, Eigen::Vector3f::UnitZ()));
    path.emplace_back(nav::core::Pose::Identity() *
                      Eigen::AngleAxisf(finalOrientation, Eigen::Vector3f::UnitZ()));

    const auto trajectory = nav::core::GlobalTrajectory::FromPath(path, 100);


    gp::OrientationOptimizer::Params params;
    gp::OrientationOptimizer optimizer(trajectory, params);

    // only the middle point should be changed.
    const gp::OrientationOptimizer::OptimizationResult result = optimizer.optimize();

    const auto toYaw = [](const nav::core::GlobalTrajectoryPoint& pt) -> double
    { return simox::math::mat4f_to_xyyaw(pt.waypoint.pose.matrix()).z(); };

    const std::vector<double> orientations =
        result.trajectory->points() | ranges::views::transform(toYaw) | ranges::to_vector;

    ARMARX_INFO << VAROUT(orientations);

    // # of points should be unchanged
    BOOST_REQUIRE_CLOSE(orientations.front(), initialOrientation, 0.01);
    BOOST_REQUIRE_CLOSE(orientations.back(), finalOrientation, 0.01);
}
