/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Visu.h"

#include <string>

#include <SimoxUtility/math/pose.h>

#include <RobotAPI/components/ArViz/Client/Layer.h>

#include <armarx/navigation/core/Graph.h>
#include <armarx/navigation/core/LocationUtils.h>
#include <armarx/navigation/core/basic_types.h>

namespace armarx::navigation::graph
{

    viz::Pose
    VertexVisu::Pose::draw(const std::string& name, const core::Pose& pose) const
    {
        return viz::Pose(name).pose(pose).scale(scale);
    }

    viz::Arrow
    VertexVisu::ForwardArrow::draw(const std::string& name, const core::Pose& pose) const
    {
        return viz::Arrow(name + " forward")
            .fromTo(
                pose.translation(),
                simox::math::transform_position(pose.matrix(), length * Eigen::Vector3f::UnitY()))
            .color(color)
            .width(width);
    }

    void
    VertexVisu::draw(viz::Layer& layer,
                     core::Graph::ConstVertex vertex,
                     visu::ObjectParserInfo info) const
    {
        const auto res = core::resolveLocation(info.objects, info.info, vertex.attrib().getPose());
        if (res.pose.has_value())
        {
            draw(layer, vertex.attrib().getName(), res.pose.value());
        }
    }

    void
    VertexVisu::draw(viz::Layer& layer, const std::string& name, const core::Pose& pose) const
    {
        if (this->pose.has_value())
        {
            layer.add(this->pose->draw(name, pose));
        }
        if (forwardArrow.has_value())
        {
            layer.add(forwardArrow->draw(name, pose));
        }
    }

    viz::Arrow
    EdgeVisu::Arrow::draw(const std::string& sourceName,
                          const std::string& targetName,
                          const core::Pose& source,
                          const core::Pose& target) const
    {
        return viz::Arrow(sourceName + " -> " + targetName)
            .fromTo(source.translation(), target.translation())
            .width(width)
            .color(color);
    }

    void
    EdgeVisu::draw(viz::Layer& layer,
                   core::Graph::ConstEdge edge,
                   visu::ObjectParserInfo info) const
    {
        const auto src =
            core::resolveLocation(info.objects, info.info, edge.source().attrib().getPose());
        const auto tar =
            core::resolveLocation(info.objects, info.info, edge.target().attrib().getPose());

        if (src.pose.has_value() && tar.pose.has_value() && arrow.has_value())
        {
            layer.add(arrow->draw(edge.source().attrib().getName(),
                                  edge.target().attrib().getName(),
                                  src.pose.value(),
                                  tar.pose.value()));
        }
    }

    void
    GraphVisu::draw(viz::Layer& layer, const core::Graph& graph, visu::ObjectParserInfo info) const
    {
        if (vertex.has_value())
        {
            for (core::Graph::ConstVertex v : graph.vertices())
            {
                vertex->draw(layer, v, info);
            }
        }
        if (edge.has_value())
        {
            for (core::Graph::ConstEdge e : graph.edges())
            {
                edge->draw(layer, e, info);
            }
        }
    }

} // namespace armarx::navigation::graph
