/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ObjectPoseClientWidget.h"

#include <memory>
#include <string>
#include <vector>

#include <QLabel>
#include <QPushButton>
#include <qboxlayout.h>
#include <qdialog.h>
#include <qobject.h>
#include <qobjectdefs.h>
#include <qwidget.h>

#include <ArmarXCore/core/ManagedIceObject.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectInfo.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectPoseClient.h>
#include <RobotAPI/libraries/ArmarXObjects/forward_declarations.h>

#include "ConnectDialog.h"

namespace armarx::navigation::qt_plugins::location_graph_editor
{

    objectposes::Connection::Connection(ObjectPoseStorageInterfacePrx objectPoseStorageComponent) :
        _objectPoseStorageComponent(
            std::make_unique<ObjectPoseStorageInterfacePrx>(objectPoseStorageComponent)),
        _objectPoseClient(std::make_unique<objpose::ObjectPoseClient>(*_objectPoseStorageComponent))
    {
        ARMARX_CHECK_NOT_NULL(_objectPoseStorageComponent);
        ARMARX_CHECK(_objectPoseClient->isConnected());
    }

    // destructor has to be defined here, because the content of the unique pointers was forward declared
    objectposes::Connection::~Connection() = default;

    armarx::ObjectPoseStorageInterfacePrx
    objectposes::Connection::getObjectPoseStorageComponent() const
    {
        ARMARX_CHECK_NOT_NULL(_objectPoseStorageComponent);
        return *_objectPoseStorageComponent;
        return nullptr;
    }

    std::string
    objectposes::Connection::getConnectedName() const
    {
        return getObjectPoseStorageComponent()->ice_getIdentity().name;
    }

    objpose::ObjectPoseClient&
    objectposes::Connection::getObjectClient()
    {
        return *_objectPoseClient;
    }

    objpose::ObjectPoseMap
    objectposes::Connection::getObjectPoseMap()
    {
        return _objectPoseClient->fetchObjectPosesAsMap();
    }

    std::vector<ObjectInfo>
    objectposes::Connection::getObjectInfo()
    {
        const auto& finder = _objectPoseClient->getObjectFinder();
        return finder.findAllObjects();
    }

    ObjectPoseClientWidget::ObjectPoseClientWidget(ManagedIceObject& component, QWidget* parent) :
        QWidget(parent), _component(component)
    {
        _statusLabel = new QLabel("Not connected", this);
        _connectButton = new QPushButton("Connect ...", this);

        QBoxLayout* layout = new QVBoxLayout(this);
        setLayout(layout);

        layout->addWidget(_statusLabel);
        layout->addWidget(_connectButton);


        connect(_connectButton, &QPushButton::pressed, this, &This::connectToObjectPoseComponent);

        connect(this,
                &This::connected,
                this,
                [this]()
                {
                    QString name = QString::fromStdString(connection().getConnectedName());
                    _statusLabel->setText("Connected to '" + name + "'");
                });
    }

    bool
    ObjectPoseClientWidget::isConnected() const
    {
        return _connection.has_value();
    }

    objectposes::Connection&
    ObjectPoseClientWidget::connection()
    {
        return _connection.value();
    }

    void
    ObjectPoseClientWidget::connectToObjectPoseComponent()
    {
        ConnectDialog<ObjectPoseStorageInterfacePrx> dialog("*ObjectMemory", _component, this);
        switch (dialog.exec())
        {
            case QDialog::Accepted:
                break;
            case QDialog::Rejected:
                return;
        }

        auto objectPoseStorageComponent = dialog.getProxy();
        if (objectPoseStorageComponent)
        {
            _connection.emplace(objectPoseStorageComponent);
            emit connected();
        }
    }

} // namespace armarx::navigation::qt_plugins::location_graph_editor
