/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Tobias Gröger ( tobias dot groeger at student dot kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE Navigation::ArmarXLibraries::human
#define ARMARX_BOOST_TEST

#include <cstdlib> /* srand, rand */
#include <ctime> /* time */

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/util/time.h>

#include <RobotAPI/libraries/ukfm/UnscentedKalmanFilter.h>

#include <armarx/navigation/Test.h>
#include <armarx/navigation/human/HumanSystemModel.h>

#include <manif/SO2.h>
#include <sciplot/sciplot.hpp>

namespace armarx::navigation::human
{
    using T = double; //TODO double or float?
    using Vector = Eigen::Matrix<T, 2, 1>;
    using SystemModelT = kalman_filter::SystemModelSO2xR2<T>;

    constexpr T RAD = 40;
    constexpr long num_timesteps = 3000;
    constexpr T max_time = 1;
    constexpr T dt = max_time / num_timesteps;
    constexpr T c = (1 / max_time) * 2 * M_PI; // rotation per time

    constexpr T rot_noise_std = 0.2;
    constexpr T pos_noise_std = 5;
    constexpr T obs_rot_noise_std = 0.2;
    constexpr T obs_pos_noise_std = 5;
    constexpr T initial_offset_angle = 1 * 10 * M_PI / 180;
    const Vector initial_offet_pos = 0.8 * RAD * Vector(1, 0.5);

    void
    simulate_trajectory(std::vector<SystemModelT::StateT>& states,
                        std::vector<SystemModelT::ControlT>& omegas)
    {
        constexpr T startAngle = 0;

        SystemModelT::StateT state;
        state.position = Vector{RAD, 0};
        state.orientation = manif::SO2<T>(startAngle);
        states.push_back(state);

        for (int i = 1; i < num_timesteps; i++)
        {
            const T t = dt * i;
            const T angle = t * c;

            const Vector last_pos = states.at(i - 1).position;
            //const Vector pos(angle, std::cos(angle));
            const Vector pos = RAD * Vector(std::cos(angle), std::sin(angle));

            BOOST_TEST(((last_pos - pos).norm() < 0.1 * RAD),
                       "position differs too much from last step: " << (last_pos - pos).norm());

            SystemModelT::ControlT control;
            //TODO (SALt): check angular velocity (>pi overflow)
            control.angular_velocity.coeffs() << c; //(last_angle - angle) / dt;
            control.euclidean_velocity << (pos - last_pos) / dt;
            //control.euclidean_velocity = Eigen::Vector2d{0, c};

            const SystemModelT::StateT propagated = SystemModelT::propagationFunction(
                state, control, SystemModelT::ControlNoiseT::Zero(), dt);

            state.position = pos;
            state.orientation = manif::SO2<T>(angle + startAngle);

            T pos_diff = (propagated.position - state.position).norm();
            BOOST_TEST(pos_diff < 2e-10,
                       "propagated position differs too much from real position: " << pos_diff);

            T rot_diff = propagated.orientation.angle() - state.orientation.angle();
            BOOST_TEST(rot_diff < 2e-10,
                       "propagated rotation differs too much from real orientation: " << rot_diff);

            states.push_back(propagated);


            // add noise
            control.angular_velocity.coeffs() += rot_noise_std * Eigen::Matrix<T, 1, 1>::Random();
            control.euclidean_velocity += pos_noise_std * Vector::Random();
            omegas.push_back(control);
        }
    }

    void
    simulate_observation(const std::vector<SystemModelT::StateT>& states,
                         std::vector<SystemModelT::ObsT>& observations)
    {
        for (const auto& state : states)
        {
            SystemModelT::ObsT obs = SystemModelT::observationFunction(state);
            SystemModelT::ObsT true_obs;
            true_obs.segment(0, 1) = state.orientation.log().coeffs();
            true_obs.segment(1, 2) = state.position;

            BOOST_TEST(
                (obs - true_obs).norm() < std::numeric_limits<T>::epsilon(),
                "observation differs too much from real observation: " << (obs - true_obs).norm());

            SystemModelT::ObsT noise = SystemModelT::ObsT::Random();
            noise.segment(0, 1) *= obs_rot_noise_std;
            noise.segment(1, 2) *= obs_pos_noise_std;

            observations.push_back(obs + noise);
        }
    }

    BOOST_AUTO_TEST_CASE(so2KalmanFilterTest)
    {
        srand(time(NULL));

        std::vector<SystemModelT::StateT> states;
        std::vector<SystemModelT::ControlT> omegas;
        std::vector<SystemModelT::ObsT> observations;
        simulate_trajectory(states, omegas);
        simulate_observation(states, observations);

        ARMARX_INFO << "Num States: " << states.size() << " Num Controls: " << omegas.size()
                    << " Num Obs: " << observations.size();

        UnscentedKalmanFilter<SystemModelT>::PropCovT Q =
            UnscentedKalmanFilter<SystemModelT>::PropCovT::Identity();
        Q.block<1, 1>(0, 0) *= rot_noise_std * rot_noise_std;
        Q.block<2, 2>(1, 1) *= pos_noise_std * pos_noise_std;

        UnscentedKalmanFilter<SystemModelT>::ObsCovT R =
            UnscentedKalmanFilter<SystemModelT>::ObsCovT::Identity();
        R.block<1, 1>(0, 0) *= obs_rot_noise_std * obs_rot_noise_std;
        R.block<2, 2>(1, 1) *= obs_pos_noise_std * obs_pos_noise_std;

        UnscentedKalmanFilter<SystemModelT>::StateCovT P0 =
            UnscentedKalmanFilter<SystemModelT>::StateCovT::Identity();
        P0.block<1, 1>(0, 0) *= initial_offset_angle * initial_offset_angle;
        P0.block<2, 2>(1, 1) *= initial_offet_pos.norm() * initial_offet_pos.norm();

        UnscentedKalmanFilter<SystemModelT>::AlphaT alpha;
        alpha(0) = 0.5; // state
        alpha(1) = 0.5; // control
        alpha(2) = 0.5; // update

        SystemModelT::StateT state0;
        state0.orientation = manif::SO2<T>(states.at(0).orientation.angle() + initial_offset_angle);
        state0.position = states.at(0).position + initial_offet_pos;

        UnscentedKalmanFilter<SystemModelT> ukf(Q, R, alpha, state0, P0);

        std::vector<SystemModelT::StateT> ukf_states;
        std::vector<UnscentedKalmanFilter<SystemModelT>::StateCovT> ukf_Ps;
        ukf_states.push_back(state0);
        ukf_Ps.push_back(P0);

        std::vector<T> x_true, y_true, x_obs, y_obs, x_ukf, y_ukf;
        std::vector<T> a_true, a_obs, a_ukf;
        std::vector<T> x_ukf_full, y_ukf_full, a_ukf_full;
        std::vector<T> ukf_pos_var, ukf_rot_var, ukf_var;

        for (int i = 1; i < num_timesteps; i++)
        {
            // propagate
            TIMING_START(LOOP);
            TIMING_START(PROPAGATION);
            ukf.propagation(omegas.at(i - 1), dt);
            TIMING_END(PROPAGATION);
            if ((i - 1) % 50 == 0)
            {
                TIMING_START(UPDATE);
                ukf.update(observations.at(i));
                TIMING_END(UPDATE);
                TIMING_START(REST);
                const SystemModelT::StateT& current_state = ukf.getCurrentState();

                BOOST_TEST_MESSAGE(
                    "Position Diff: " << (states.at(i).position - current_state.position).norm());
                BOOST_TEST_MESSAGE(
                    "Orientation Diff: "
                    << (states.at(i).orientation - current_state.orientation).coeffs().norm());

                BOOST_TEST_MESSAGE("Max Cov "
                                   << ukf.getCurrentStateCovariance().cwiseAbs().maxCoeff());
                BOOST_TEST_MESSAGE("Diag: " << (ukf.getCurrentStateCovariance() -
                                                Eigen::Matrix<T, 3, 3>::Identity())
                                                   .norm());

                x_obs.push_back(observations.at(i)(1));
                y_obs.push_back(observations.at(i)(2));
                a_obs.push_back(observations.at(i)(0));

                x_ukf.push_back(ukf.getCurrentState().position.x());
                y_ukf.push_back(ukf.getCurrentState().position.y());
                a_ukf.push_back(ukf.getCurrentState().orientation.log().coeffs()(0));

                TIMING_END(REST);
            }
            x_true.push_back(states.at(i).position.x());
            y_true.push_back(states.at(i).position.y());
            a_true.push_back(states.at(i).orientation.log().coeffs()(0));

            x_ukf_full.push_back(ukf.getCurrentState().position.x());
            y_ukf_full.push_back(ukf.getCurrentState().position.y());
            a_ukf_full.push_back(ukf.getCurrentState().orientation.log().coeffs()(0));

            ukf_states.push_back(ukf.getCurrentState());
            ukf_Ps.push_back(ukf.getCurrentStateCovariance());

            ukf_pos_var.push_back(
                ukf.getCurrentStateCovariance().block<2, 2>(1, 1).cwiseAbs().maxCoeff());
            ukf_rot_var.push_back(
                ukf.getCurrentStateCovariance().block<1, 1>(0, 0).cwiseAbs().maxCoeff());
            ukf_var.push_back(ukf.getCurrentStateCovariance().cwiseAbs().maxCoeff());

            TIMING_END(LOOP);
        }

        sciplot::Plot2D pos_plot;
        pos_plot.xlabel("x");
        pos_plot.ylabel("y");
        pos_plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);
        pos_plot.xrange(-RAD * 1.1, RAD * 1.1).yrange(-RAD * 1.1, RAD * 1.1);
        //pos_plot.xrange(-4, 4).yrange(-4, 4);

        pos_plot.drawCurve(x_true, y_true).label("True");
        pos_plot.drawCurve(x_obs, y_obs).label("Obs");
        pos_plot.drawCurveWithPoints(x_ukf, y_ukf).label("UKF");
        pos_plot.drawCurve(x_ukf_full, y_ukf_full).label("Full UKF");


        sciplot::Plot2D orientation_plot;
        orientation_plot.xlabel("x");
        orientation_plot.ylabel("y");
        orientation_plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);
        orientation_plot.xrange(-0.1, 1.1).yrange(-4, 4);

        sciplot::Vec x = sciplot::linspace(0.0, 1.0, a_obs.size());
        sciplot::Vec x_t = sciplot::linspace(0.0, 1.0, a_true.size());
        orientation_plot.drawCurve(x_t, a_true).label("True").lineWidth(1);
        orientation_plot.drawCurve(x, a_obs).label("Obs").lineWidth(1);
        orientation_plot.drawCurve(x, a_ukf).label("UKF").lineWidth(1);
        orientation_plot.drawCurve(x_t, a_ukf_full).label("Full UKF").lineWidth(1);


        sciplot::Plot2D error_plot;
        error_plot.xlabel("x");
        error_plot.ylabel("y");
        error_plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);
        error_plot.xrange(-0.1, 1.1).yrange("*", "*");

        error_plot.drawCurve(x_t, ukf_pos_var).label("Pos variance").lineWidth(1);
        error_plot.drawCurve(x_t, ukf_rot_var).label("Rot variance").lineWidth(1);
        error_plot.drawCurve(x_t, ukf_var).label("Total variance").lineWidth(1);


        // Create figure to hold plot
        sciplot::Figure fig = {{pos_plot, orientation_plot}, {error_plot}};
        // Create canvas to hold figure
        sciplot::Canvas canvas = {{fig}};
        canvas.size(1000, 1000);

        // Show the plot in a pop-up window
        canvas.show();

        // Save the plot to a PDF file
        canvas.save("so2_kalman_output.pdf");
    }

} // namespace armarx::navigation::human
