/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <map>
#include <optional>
#include <ostream>
#include <string>
#include <vector>

#include <RobotAPI/libraries/armem/client/Reader.h>
#include <RobotAPI/libraries/armem/client/util/SimpleReaderBase.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem_locations/aron/Location.aron.generated.h>

#include <armarx/navigation/core/Graph.h>
#include <armarx/navigation/core/location.h>

namespace armarx::navigation::memory::client::graph
{

    struct ResolvedLocationName
    {
        armarx::armem::MemoryID locationId;
        std::string locationName;
        double matchRatio;
    };

    struct ResolveLocationNameResult
    {
        std::optional<ResolvedLocationName> resolved;
        std::optional<std::string> errorMessage;
    };

    class Reader : virtual public armem::client::util::SimpleReaderBase
    {
    public:
        using armem::client::util::SimpleReaderBase::SimpleReaderBase;

        std::string propertyPrefix() const override;
        Properties defaultProperties() const override;

        std::map<std::string, core::Location> locations();

        std::vector<armarx::navigation::core::Graph> graphs();

        core::Location resolveLocationId(const std::string& locationId);


        /**
         * @brief Resolve a location by human-readable name.
         *
         * Queries location names from the navigation memory, matches the given  name to the
         * locations' names, and returns the ID of the location with the best match.
         *
         * @param locationName The human-readable name.
         * @param minMatchRatio The minimum match ratio required to accept a match.
         * @param log Optional out-stream for logging about the matching.
         * @return Resolved location name or error message.
         */
        ResolveLocationNameResult resolveLocationName(const std::string& locationName,
                                                      double minMatchRatio,
                                                      std::ostream* log = nullptr);
        /**
         * @brief Resolve a location by human-readable name.
         *
         * Matches the given  name to the locations' names, and returns the ID of the location with
         * the best match.
         *
         * @param locations Locations, e.g. as returned by `Reader::queryLocations()`.
         * @param locationName The human-readable name.
         * @param minMatchRatio The minimum match ratio required to accept a match.
         * @param log Optional out-stream for logging about the matching.
         * @return Resolved location name or error message.
         */
        static ResolveLocationNameResult
        ResolveLocationName(const std::map<armem::MemoryID, location::arondto::Location>& locations,
                            const std::string& locationName,
                            double minMatchRatio,
                            std::ostream* log = nullptr);

        void connect(armarx::armem::client::MemoryNameSystem& mns) override;


    protected:
        std::map<armem::MemoryID, location::arondto::Location> queryLocations();

        armem::wm::CoreSegment allLocations();
        armem::wm::CoreSegment allGraphs();

    private:
        armem::client::Reader memoryReaderLocations;
        armem::client::Reader memoryReaderGraphs;
    };

} // namespace armarx::navigation::memory::client::graph
