/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <optional>

#include <SimoxUtility/color/Color.h>
#include <SimoxUtility/color/ColorMap.h>

#include <armarx/navigation/core/basic_types.h>
#include <armarx/navigation/core/types.h>
#include <armarx/navigation/util/geometry.h>

#include "SafetyGuard.h"

namespace armarx::viz
{
    struct Layer;
}

namespace armarx::navigation::safety_guard
{

    struct ProximityFieldParams
    {
        bool reduceVelocity{true};

        float minDistance{500.F};
        float maxDistance{2000.F};

        // Parameters to reduce the velocity in proximity to an obstacle between minDistance and maxDistance.
        // It is v = v_max / (1 + lambda * d_s^lambda) with d_s = (1 - (d / maxDistance))^k
        float k{4.F};
        float lambda{6.F};
    };

    struct LaserBasedProximityParams : public SafetyGuardParams
    {
        float robotRadius{500.F}; //TODO(utetg): should there be a default, if yes what

        bool enableHumans{true};
        bool enableLaserScanners{true};

        core::TwistLimits maxVelocity{
            .linear = 500.F,
            .angular = 0.5F,
        };

        ProximityFieldParams humanProximityField;
        ProximityFieldParams laserScannerProximityField;

        Algorithms algorithm() const override;
        aron::data::DictPtr toAron() const override;
        static LaserBasedProximityParams FromAron(const aron::data::DictPtr& dict);
    };

    class LaserBasedProximity : virtual public SafetyGuard
    {
    public:
        using Params = LaserBasedProximityParams;

        LaserBasedProximity(const Params& params, const core::Scene& scene, const Context& ctx);
        ~LaserBasedProximity() override = default;

        SafetyGuardResult computeSafetyLimits() override;

    private:
        std::optional<core::TwistLimits> safetyLimitsHumans(viz::Layer& layer) const;
        std::optional<core::TwistLimits> safetyLimitsLaserScanners(viz::Layer& layer) const;

        std::pair<float, Eigen::Vector2f>
        calculateMinDistance(const util::geometry::polygon_type& convexHull,
                             const core::Pose& robot_T_global,
                             const std::vector<Eigen::Vector3f>& globalPoints) const;

        std::pair<core::TwistLimits, float>
        evaluateProximityField(const ProximityFieldParams& proximityField, float minDistance) const;

    protected:
        const Params params;

    private:
        simox::ColorMap humanColorMap_;
        simox::ColorMap laserColorMap_;
    };
} // namespace armarx::navigation::safety_guard
