#include "NavigateTo.h"

#include <chrono>
#include <future>
#include <string>

#include <Eigen/Core>
#include <Eigen/Geometry>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/util/StringHelpers.h>

#include <RobotAPI/libraries/skills/core/Skill.h>
#include <RobotAPI/libraries/skills/core/SkillStatusUpdate.h>

#include <armarx/navigation/client/NavigationStackConfig.h>
#include <armarx/navigation/client/Navigator.h>
#include <armarx/navigation/client/services/IceNavigator.h>
#include <armarx/navigation/core/types.h>
#include <armarx/navigation/global_planning/SPFA.h>
#include <armarx/navigation/trajectory_control/local/TrajectoryFollowingController.h>

namespace armarx::navigation::skills
{
    template <typename R>
    bool
    is_ready(std::future<R> const& f)
    {
        if (!f.valid())
        {
            return true;
        }
        std::future_status status = f.wait_for(std::chrono::seconds(0));
        if (status == std::future_status::ready)
        {
            return true;
        }
        else if (status == std::future_status::deferred)
        {
            // if the task has not been started --> start the task
            f.wait();
            return true;
        }
        else
        {
            return false;
        }
    }

    NavigateTo::NavigateTo(const Services& srv) : Base(DefaultSkillDescription())
    {
        srv_.emplace(srv);
    }

    ::armarx::skills::Skill::InitResult
    NavigateTo::init(const Base::SpecializedInitInput& in)
    {

        armarx::navigation::traj_ctrl::local::TrajectoryFollowingControllerParams
            trajControllerParams{};

        client::GeneralConfig generalParams{};

        if (in.parameters.velocityLimitLinear.has_value())
        {
            trajControllerParams.limits.linear = in.parameters.velocityLimitLinear.value();
            generalParams.maxVel.linear = in.parameters.velocityLimitLinear.value();
        }

        if (in.parameters.velocityLimitAngular.has_value())
        {
            trajControllerParams.limits.angular = in.parameters.velocityLimitAngular.value();
            generalParams.maxVel.angular = in.parameters.velocityLimitAngular.value();
        }

        ARMARX_INFO << "Configuring navigator to respect the velocity limits:";
        ARMARX_INFO << "linear: " << trajControllerParams.limits.linear;
        ARMARX_INFO << "angular: " << trajControllerParams.limits.angular;

        // parameterize the navigation stack
        client::NavigationStackConfig cfg;
        cfg.general(generalParams);
        cfg.globalPlanner(armarx::navigation::global_planning::SPFAParams{});
        cfg.trajectoryController(trajControllerParams);

        const std::string configId = DefaultSkillDescription().skillId.skillName;

        // configure the `navigator` which provides a simplified and typed interface to the navigation server
        memorySubscriber.emplace(configId, srv_->memoryNameSystem);

        // register our config
        ARMARX_INFO << "Registering config";
        srv_->iceNavigator.createConfig(cfg, configId);

        navigator.emplace(client::Navigator::InjectedServices{
            .navigator = &srv_->iceNavigator, .subscriber = &memorySubscriber.value()});

        return ::armarx::skills::Skill::InitResult{
            .status = ::armarx::skills::TerminatedSkillStatus::Succeeded};
    }

    ::armarx::skills::Skill::MainResult
    NavigateTo::main(const Base::SpecializedMainInput& in)
    {
        const Eigen::Isometry3f globalTarget(in.parameters.targetPose);

        ARMARX_INFO << "moving to target " << VAROUT(globalTarget.matrix());

        // execute
        ARMARX_INFO << "Sending navigation request";
        navigator->moveTo(globalTarget, core::NavigationFrame::Absolute);

        // Wait until goal is reached
        ARMARX_INFO << "Waiting until goal is reached.";

        auto future = std::async(std::launch::async, [this]() { return navigator->waitForStop(); });
        while (not is_ready(future))
        {
            if (shouldSkillTerminate())
            {
                ARMARX_INFO << "Skill was aborted by user " << stopped << ", " << timeoutReached;
                navigator->stop();
                break;
            }
        }

        auto se = future.get();
        if (se)
        {
            ARMARX_INFO << "Goal `" << Eigen::Isometry3f(in.parameters.targetPose).translation()
                        << "`reached.";
        }
        else
        {
            if (se.isSafetyStopTriggeredEvent())
            {
                ARMARX_ERROR << "Safety stop was triggered!";

                return ::armarx::skills::Skill::MainResult{
                    .status = ::armarx::skills::TerminatedSkillStatus::Failed};
            }
            if (se.isUserAbortTriggeredEvent())
            {
                ARMARX_ERROR << "Aborted by user!";

                return ::armarx::skills::Skill::MainResult{
                    .status = ::armarx::skills::TerminatedSkillStatus::Failed};
            }
            if (se.isInternalErrorEvent())
            {
                ARMARX_ERROR << "Unknown internal error occured! "
                             << se.toInternalErrorEvent().message;

                return ::armarx::skills::Skill::MainResult{
                    .status = ::armarx::skills::TerminatedSkillStatus::Failed};
            }
        }


        return ::armarx::skills::Skill::MainResult{
            .status = ::armarx::skills::TerminatedSkillStatus::Succeeded};
    }

    void
    NavigateTo::onStopRequested()
    {
        if (navigator.has_value())
        {
            ARMARX_CHECK(navigator.has_value());
            navigator->stop();
        }
    }


} // namespace armarx::navigation::skills
