/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    navigation::NavigationGroup
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NavigateRelative.h"

#include <RobotAPI/components/armem/MemoryNameSystem/MemoryNameSystem.h>
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/core/FramedPose.h>

#include <armarx/navigation/client/NavigationStackConfig.h>
#include <armarx/navigation/client/Navigator.h>
#include <armarx/navigation/client/PathBuilder.h>
#include <armarx/navigation/client/services/IceNavigator.h>
#include <armarx/navigation/client/services/MemoryPolling.h>
#include <armarx/navigation/client/services/SimpleEventHandler.h>
#include <armarx/navigation/client/types.h>
#include <armarx/navigation/global_planning/Point2Point.h>
#include <armarx/navigation/trajectory_control/local/TrajectoryFollowingController.h>

namespace armarx::navigation::statecharts::navigation_group
{
    // DO NOT EDIT NEXT LINE
    NavigateRelative::SubClassRegistry
        NavigateRelative::Registry(NavigateRelative::GetName(), &NavigateRelative::CreateInstance);

    void
    NavigateRelative::onEnter()
    {
        // put your user code for the enter-point here
        // execution time should be short (<100ms)
    }

    void
    NavigateRelative::run()
    {
        const core::Pose relativeTarget(in.getrelativePose()->toEigen());

        ARMARX_INFO << "moving to relative target " << VAROUT(relativeTarget.matrix());

        // parameterize the navigation stack
        client::NavigationStackConfig cfg;
        cfg.general(client::GeneralConfig{});
        cfg.globalPlanner(armarx::navigation::global_planning::Point2PointParams{});
        cfg.trajectoryController(
            armarx::navigation::traj_ctrl::local::TrajectoryFollowingControllerParams{});

        // configure the `navigator` which provides a simplified and typed interface to the navigation server
        client::IceNavigator iceNavigator(getNavigator());

        armem::client::MemoryNameSystem mns(getMemoryNameSystem());

        client::MemoryPolling memoryEventPolling(GetName(), mns);

        // register our config
        ARMARX_INFO << "Registering config";
        iceNavigator.createConfig(cfg, GetName());

        client::Navigator navigator{client::Navigator::InjectedServices{
            .navigator = &iceNavigator, .subscriber = &memoryEventPolling}};

        // execute
        ARMARX_INFO << "Sending navigation request";
        navigator.moveTo(relativeTarget, core::NavigationFrame::Relative);

        // Wait until goal is reached
        ARMARX_INFO << "Waiting until goal is reached.";
        client::StopEvent se = navigator.waitForStop();
        if (se)
        {
            ARMARX_INFO << "Goal `" << relativeTarget.matrix() << "`reached.";
        }
        else
        {
            if (se.isSafetyStopTriggeredEvent())
            {
                ARMARX_ERROR << "Safety stop was triggered!";
                emitFailure();
            }
            else if (se.isUserAbortTriggeredEvent())
            {
                ARMARX_ERROR << "Aborted by user!";
                emitFailure();
            }
            else if (se.isInternalErrorEvent())
            {
                ARMARX_ERROR << "Unknown internal error occured! "
                             << se.toInternalErrorEvent().message;
                emitFailure();
            }
        }

        emitSuccess();
    }

    //void NavigateRelative::onBreak()
    //{
    //    // put your user code for the breaking point here
    //    // execution time should be short (<100ms)
    //}

    void
    NavigateRelative::onExit()
    {
        // put your user code for the exit point here
        // execution time should be short (<100ms)
    }

    // DO NOT EDIT NEXT FUNCTION
    XMLStateFactoryBasePtr
    NavigateRelative::CreateInstance(XMLStateConstructorParams stateData)
    {
        return {new NavigateRelative(stateData)};
    }
} // namespace armarx::navigation::statecharts::navigation_group
